/*
 * Kernel module for TQMP kernel mode media proxying.
 *
 * Shall work on post 4.18 kernels, where NAT API changed a lot (very late Deb9 and Deb10).
 *
 * Copyright (c) 2000-2020 DEVERTO Ltd.
 */

#include <linux/module.h>
#include <net/ip.h>
#include <net/netfilter/nf_conntrack_core.h>
#include <net/netfilter/nf_nat_l3proto.h>

MODULE_AUTHOR("Deverto Ltd <admin@deverto.com>");
MODULE_DESCRIPTION("Deverto TQMP kernel module");
MODULE_LICENSE("GPL"); // nf_ct_netns_get (and others) need this
MODULE_ALIAS("dvrt_tqmp");

// these shall be in sync with TQMP
#define FORWARDMASK 0xffff
#define REVERSEMASK 0x7fff
#define REVERSESHIFT 16

#define MARKOFFSET 17
#define MODULENAME "nf_dvrt_tqmp"

#if 0
#define CREATE_IPV4_BYTES(ip) ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff
#endif

// copy conntrack mark's appropriate part to packet mark
unsigned int mark(void* priv, struct sk_buff* skb, const struct nf_hook_state* state)
{
    const struct iphdr* iph;
    const struct udphdr* udph;
    const struct tcphdr* tcph;
    const struct nf_conn* ct;
    enum ip_conntrack_info ctinfo;
    const struct nf_conntrack_tuple* ct_odir_tuple;

    // only IPv4 is processed here (no marking needed for IPv6)
    if (skb->protocol != htons(ETH_P_IP))
	return NF_ACCEPT;

    // only UDP and TCP is processed
    iph = ip_hdr(skb);
    if (iph->protocol != IPPROTO_UDP && iph->protocol != IPPROTO_TCP)
	return NF_ACCEPT;

    // packet not in conntrack
    ct = nf_ct_get(skb, &ctinfo);
    if (!ct)
	return NF_ACCEPT;

    // TODO: remove if TQMP will set conntrack mark=0 for net->net0 traffic
    // conntrack not created by TQMP
    if (ct->mark == 0)
	return NF_ACCEPT;

    // Need to compare src and dst ip:port to decide direction, because (due to overlapping IP)
    // dst IP:PORT for original and reply dir can be the same,
    // if TQT IP on both legs is the same, and MP chose the same PORT on both legs.
    // (Consequently dest IP:PORT is the same for both directions.)
    //
    // Order of comparison choosen for best performance
    // (src port tend to differ the most for the 2 directions, dest IP the least).
    ct_odir_tuple = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
    if (iph->protocol == IPPROTO_UDP)
    {
	udph = (struct udphdr*)((void*)iph + ip_hdrlen(skb));

#if 0
	pr_info("udp packet %d.%d.%d.%d:%d -> %d.%d.%d.%d:%d mark: 0x%04x ctstat: 0x%lx\n",
		CREATE_IPV4_BYTES(ct_odir_tuple->src.u3.ip),
		htons(ct_odir_tuple->src.u.udp.port),
		CREATE_IPV4_BYTES(ct_odir_tuple->dst.u3.ip),
		htons(ct_odir_tuple->dst.u.udp.port),
		skb->mark, ct->status);
#endif

	if (ct_odir_tuple->src.u.udp.port == udph->source &&
		ct_odir_tuple->src.u3.ip == iph->saddr &&
		ct_odir_tuple->dst.u.udp.port == udph->dest &&
		ct_odir_tuple->dst.u3.ip == iph->daddr)
	    skb->mark = ct->mark & FORWARDMASK;
	else
	    skb->mark = (ct->mark >> REVERSESHIFT) & REVERSEMASK;
    }
    else if (iph->protocol == IPPROTO_TCP)
    {
	tcph = (struct tcphdr*)((void*)iph + ip_hdrlen(skb));
	if (ct_odir_tuple->src.u.tcp.port == tcph->source &&
		ct_odir_tuple->src.u3.ip == iph->saddr &&
		ct_odir_tuple->dst.u.tcp.port == tcph->dest &&
		ct_odir_tuple->dst.u3.ip == iph->daddr)
	    skb->mark = ct->mark & FORWARDMASK;
	else
	    skb->mark = (ct->mark >> REVERSESHIFT) & REVERSEMASK;
    }
    else
	return NF_ACCEPT;

    // add a big number to packet marker (to avoid collision with used ones)
    skb->mark |= 1 << MARKOFFSET;

    return NF_ACCEPT;
}

// needed to start nat, does nothing
// (Note: this function is actually only called for new conntrack entries and if a packet has no NAT info already set.
// This practically means that it is called very rarely.)
unsigned int dummynat( void* priv, struct sk_buff* skb, const struct nf_hook_state* state)
{
    return NF_ACCEPT;
}

// which netfilter hook to register
static const struct nf_hook_ops markho =
{
    .hook = mark,
    .hooknum = NF_INET_PRE_ROUTING,
    .pf = NFPROTO_IPV4,
    .priority = NF_IP_PRI_CONNTRACK + 1 // right after conntrack
};


// needed to start nat
static const struct nf_hook_ops natho =
{
    .hook = dummynat,
    .hooknum = NF_INET_PRE_ROUTING,
    .pf = NFPROTO_IPV4,
    .priority = NF_IP_PRI_NAT_DST + 1 // let all other nat function go before, we don't want to do anything
};

static int __init load(void)
{
    int ret = 0;

    // start conntrack (register as conntrack user for IPV4)
    ret = nf_ct_netns_get(&init_net, NFPROTO_IPV4);
    if (ret)
    {
	pr_err("%s: module stop, cannot enable conntrack support, error: %d\n", MODULENAME, ret);
	return ret;
    }

    // start IPv4 NAT (start nf_nat_ipv4, aka nf_nat_l3proto_ipv4)
    // Note: without doing nf_nat_l3proto_ipv4_register_fn, nf_nat_ipv4 has no user
    // and does not process NAT info in conntrack
    // (till someone else registers as its user, eg. nftables by issuing some 'iptables -t nat ...' command).
    // Same true for IPv6.
    ret = nf_nat_l3proto_ipv4_register_fn(&init_net, &natho);
    if (ret)
    {
	nf_ct_netns_put(&init_net, NFPROTO_IPV4);
	pr_err("%s: module stop, cannot enable IPv4 NAT support, error: %d\n", MODULENAME, ret);
	return ret;
    }

    // start mark manipulation
    ret = nf_register_net_hook(&init_net, &markho);
    if (ret)
    {
	nf_nat_l3proto_ipv4_unregister_fn(&init_net, &natho);
	nf_ct_netns_put(&init_net, NFPROTO_IPV4);
	pr_err("%s: module stop, cannot start mark manipulation, error: %d\n", MODULENAME, ret);
	return ret;
    }
    pr_info("%s: module started\n", MODULENAME);
    return ret;
}

static void __exit remove(void)
{
    nf_nat_l3proto_ipv4_unregister_fn(&init_net, &natho);
    nf_unregister_net_hook(&init_net, &markho);
    nf_ct_netns_put(&init_net, NFPROTO_IPV4);
    pr_info("%s: module stopped\n", MODULENAME);
}

module_init(load);
module_exit(remove);
